/*

 Copyright (C) 2015 AG-KW, Franke & Schwarz GbR <zentrale@ag-kw.de>

 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software Foundation;
 either version 3 of the License, or (at your option) any later version. 

 This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 See the GNU General Public License for more details. 

 For the GNU General Public License see <http://www.gnu.org/licenses/>.
 
 Used open source libraries:
 Arduino MIDI Library v4.2: Copyright (c) 2014, Francois Best, GPL Forty Seven Effects (https://github.com/FortySevenEffects/arduino_midi_library)
 EEPROM library: Copyright (c) 2006, David A. Mellis (http://arduino.cc/en/Reference/EEPROM)
 SPI library: Copyright (c) 2010, Cristian Maglie (http://arduino.cc/en/Reference/SPI)
 
 
 KEEP IT OPEN 4LIFE!
 
 */

//******************************************************************************************************************************

#include <MIDI.h> //includes Midi-Library
MIDI_CREATE_DEFAULT_INSTANCE();

#include <EEPROM.h>
#include <SPI.h>
// sck @digitalPin #13; //SPI clock - defined by SPI library
// miso @digitalPin #12; //SPI master in / slave out - defined by SPI library
// mosi @digitalPin #11; //SPI master out / slave in - defined by SPI library

//******************************************************************************************************************************



//declaration of variables******************************************************************************************************

//pins
const byte functionbutton = 4; //function-pushbutton @ pin4
const byte pb2 = 5; //pushbutton #2 @ pin5
const byte function_led = 6; //status led @ pin6
const byte s0 = 7; const byte s1 = 8; const byte s2 = 9; //pins for addressing multiplexer 4051 (address lines)
const byte ss = 10; //SPI slave select - used as Load Pin (PL) @74HC165

//several variables
byte bit0 = 0; byte bit1 = 0; byte bit2 = 0; //bit 0 to 2 of a decimal number

byte midicc_number = 0;
byte midicc_channel = 0;

byte buttonAssign[6][8] = {
{49, 49, 49, 49, 49, 49, 49, 49},
{50, 50, 50, 50, 50, 50, 50, 50},
{52, 52, 52, 52, 52, 52, 52, 52},
{53, 53, 53, 53, 53, 53, 53, 53},
{51, 51, 51, 51, 51, 51, 51, 51},
{54, 54, 54, 54, 54, 54, 54, 54}};

byte buttonChout[6][8] = { //button midi output channels
{1, 2, 3, 4, 5, 6, 7, 8}, //button#1 to #8 on page#1
{1, 2, 3, 4, 5, 6, 7, 8}, //button#1 to #8 on page#2
{1, 2, 3, 4, 5, 6, 7, 8}, //button#1 to #8 on page#3
{1, 2, 3, 4, 5, 6, 7, 8}, //button#1 to #8 on page#4
{1, 2, 3, 4, 5, 6, 7, 8}, //button#1 to #8 on page#5
{1, 2, 3, 4, 5, 6, 7, 8}}; //button#1 to #8 on page#6

byte buttonLow[6][8] ={
{0, 0, 0, 0, 0, 0, 0, 0},
{0, 0, 0, 0, 0, 0, 0, 0}, 
{0, 0, 0, 0, 0, 0, 0, 0}, 
{0, 0, 0, 0, 0, 0, 0, 0}, 
{0, 0, 0, 0, 0, 0, 0, 0}, 
{0, 0, 0, 0, 0, 0, 0, 0}};

byte buttonHigh[6][8] ={
{1, 1, 1, 1, 1, 1, 1, 1},
{1, 1, 1, 1, 1, 1, 1, 1}, 
{1, 1, 1, 1, 1, 1, 1, 1}, 
{1, 1, 1, 1, 1, 1, 1, 1}, 
{1, 1, 1, 1, 1, 1, 1, 1}, 
{1, 1, 1, 1, 1, 1, 1, 1}};

boolean buttonBool[6][8] ={
{0, 0, 0, 0, 0, 0, 0, 0},
{0, 0, 0, 0, 0, 0, 0, 0}, 
{0, 0, 0, 0, 0, 0, 0, 0}, 
{0, 0, 0, 0, 0, 0, 0, 0}, 
{0, 0, 0, 0, 0, 0, 0, 0}, 
{0, 0, 0, 0, 0, 0, 0, 0}};

byte eightbuttons = 0; //byte of the SPI input of the current states of the eight buttons (when no button is pressed = 00000000, when all are pressed = 11111111)
byte eightbuttonsLast = 0; //byte of the last states of the eight buttons

byte faderAssign[8][8] = { //fader midi cc# assignments
{25, 25, 25, 25, 25, 25, 25, 25}, //fader#1 to #8 on page#1
{39, 39, 39, 39, 39, 39, 39, 39}, //fader#1 to #8 on page#2
{16, 16, 16, 16, 16, 16, 16, 16}, //fader#1 to #8 on page#3
{20, 20, 20, 20, 20, 20, 20, 20}, //fader#1 to #8 on page#4
{21, 21, 21, 21, 21, 21, 21, 21}, //fader#1 to #8 on page#5
{40, 40, 40, 40, 40, 40, 40, 40}, //fader#1 to #8 on page#6
{41, 41, 41, 41, 41, 41, 41, 41}, //fader#1 to #8 on page#7
{45, 45, 45, 45, 45, 45, 45, 45}}; //fader#1 to #8 on page#8

byte faderChout[8][8] = { //fader midi output channels
{1, 2, 3, 4, 5, 6, 7, 8}, //fader#1 to #8 on page#1
{1, 2, 3, 4, 5, 6, 7, 8}, //fader#1 to #8 on page#2
{1, 2, 3, 4, 5, 6, 7, 8}, //fader#1 to #8 on page#3
{1, 2, 3, 4, 5, 6, 7, 8}, //fader#1 to #8 on page#4
{1, 2, 3, 4, 5, 6, 7, 8}, //fader#1 to #8 on page#5
{1, 2, 3, 4, 5, 6, 7, 8}, //fader#1 to #8 on page#6
{1, 2, 3, 4, 5, 6, 7, 8}, //fader#1 to #8 on page#7
{1, 2, 3, 4, 5, 6, 7, 8}}; //fader#1 to #8 on page#8

byte faderValue[8][8] ={
{0, 0, 0, 0, 0, 0, 0, 0},
{0, 0, 0, 0, 0, 0, 0, 0}, 
{0, 0, 0, 0, 0, 0, 0, 0}, 
{0, 0, 0, 0, 0, 0, 0, 0}, 
{0, 0, 0, 0, 0, 0, 0, 0}, 
{0, 0, 0, 0, 0, 0, 0, 0}, 
{0, 0, 0, 0, 0, 0, 0, 0}, 
{0, 0, 0, 0, 0, 0, 0, 0}};

byte faderLow[8][8] ={
{0, 0, 0, 0, 0, 0, 0, 0},
{0, 0, 0, 0, 0, 0, 0, 0},
{0, 0, 0, 0, 0, 0, 0, 0},
{0, 0, 0, 0, 0, 0, 0, 0},
{0, 0, 0, 0, 0, 0, 0, 0},
{0, 0, 0, 0, 0, 0, 0, 0},
{0, 0, 0, 0, 0, 0, 0, 0},
{0, 0, 0, 0, 0, 0, 0, 0}};

byte faderHigh[8][8] ={
{64, 64, 64, 64, 64, 64, 64, 64},
{64, 64, 64, 64, 64, 64, 64, 64},
{64, 64, 64, 64, 64, 64, 64, 64},
{64, 64, 64, 64, 64, 64, 64, 64},
{64, 64, 64, 64, 64, 64, 64, 64},
{64, 64, 64, 64, 64, 64, 64, 64},
{64, 64, 64, 64, 64, 64, 64, 64},
{64, 64, 64, 64, 64, 64, 64, 64}};

byte faderpage = 0;
byte fader[8] = {0, 0, 0, 0, 0, 0, 0, 0}; //array: fader 1 to 8
byte faderLast[8] = {0, 0, 0, 0, 0, 0, 0, 0}; //array:  value to compare with current fader position
boolean faderAssignState[8] = {1, 0, 0, 0, 0, 0, 0, 0};

boolean in_setup = 0;
boolean function_led_state = 0;
boolean setup_feedback = 0;
boolean functionbutton_state = 0;

byte a = 0; //array pointer
byte b = 0; //array pointer
byte c = 0; //array pointer
byte e = 0; //array pointer
boolean d[8] = {0, 0, 0, 0, 0, 0, 0, 0}; //array pointer
byte ledrow = B00000000;
byte menu = 0;

//variables for timer
unsigned long t1 = 0; //t1 for button debouncing & blink without delay
unsigned long t2 = 0; //t2 for button debouncing & blink without delay

//******************************************************************************************************************************



//setup*************************************************************************************************************************

void setup() {
  pinMode(function_led, OUTPUT);  pinMode(s0, OUTPUT); pinMode(s1, OUTPUT); pinMode(s2, OUTPUT); //setting pins to output (pins for addressing multiplexer)
  pinMode(functionbutton, INPUT); //setting pin to input (functionbutton)
  pinMode(ss, OUTPUT);//ss = pin to enable/latch/load shift register
  MIDI.begin(MIDI_CHANNEL_OMNI); //initialize Midi Library with default settings & midi input channel omnimode

  //start & configure SPI
  SPI.begin(); //fixed defined: MOSI @Pin11, MISO @ Pin12, SCK @ Pin13
  SPI.setDataMode(SPI_MODE0);
  SPI.setBitOrder(MSBFIRST);
  SPI.setClockDivider(SPI_CLOCK_DIV2);
  
  MIDI.setHandleControlChange(sendrequest);
 
  for (a=0; a <= 7; a++){
    input_eightfader();
    faderLast[a] = fader[a];
  }
  a = 0;
   
  digitalWrite(function_led, LOW);
  menu = 0;
  MIDI.sendControlChange(61, 0, 1);

}

//******************************************************************************************************************************





//main program******************************************************************************************************************

void loop() {
  MIDI.read();
  
  t1=millis(); //t1 for button debouncing
  input_eightfader(); //receiving the fader positions
  output_eightfader(); //sending ctrl changes (i.e. track level) via midi
  
  digitalWrite(ss, 1);
  input_eightbuttons(); //receiving the eight buttons
  output_ledrow();
  digitalWrite(ss, 0);
  
  output_eightbuttons(); //sending ctrl changes (i.e. mute) via midi
  faderpage_select();  
  menu_eightbuttons(); //changing button functionalities (i.e. mute, solo, recarm, ...)
  
  arraypointer(); //the arraypointer - used in several functions

}

//******************************************************************************************************************************




//functions*********************************************************************************************************************

//receiving feedback from Octatrack (callback function - is called when midi control change on midi-in is received)
void sendrequest(byte channel, byte number, byte value){

  midicc_number = number;
  midicc_channel = channel;
  
  for (b = 0; b <= 5; b++){ //b is the row a.k.a. button/fader-page
    for (e = 0; e <= 7; e++){ //e is the column a.k.a. button/fader-number
      if(number == buttonAssign[b][e] && channel == buttonChout[b][e]){
        if(buttonLow[b][e] == value){
          buttonBool[b][e] = 0;
        }
        if(buttonHigh[b][e] == value){
          buttonBool[b][e] = 1;
        }
      }
      
      if(number == faderAssign[b][e] && channel == faderChout[b][e] && faderValue[b][e] != value){
        faderValue[b][e] = value;
        d[e] = 0;
      }
      
    }
  }
}


//receiving the fader positions via multiplexer 4051 (connected to analog input pin 0)
void input_eightfader(){  
  bit0 = bitRead(a,0); bit1 = bitRead(a,1); bit2 = bitRead(a,2); //addressing the multiplexer: read bits of decimals from back to front (LSB) = converts decimals (0 to 7 = 2^3) to bits
  digitalWrite(s0, bit0); digitalWrite(s1, bit1); digitalWrite(s2, bit2); //addressing the multiplexer: send bits to multiplexer 4051 for addressing
  //fader[a]=analogRead(0)>>3; //reads potentiometer positions (multiplexer 4051 connected to analog input pin 0)
  //fader[a]=map(fader[a], 0, 127, faderLow[faderpage][a], faderHigh[faderpage][a]);
  fader[a]=map(analogRead(0)>>3, 0, 127, faderLow[faderpage][a], faderHigh[faderpage][a]);
}


//sending ctrl changes (i.e. track level) via midi
void output_eightfader(){

  if(fader[a] == faderValue[faderpage][a]) {d[a]=1;}//abholen
  
  if(fader[a] != faderValue[faderpage][a] && d[a] == 1) { //compares current potentiometer position with  "recorded" value & sends only if actual potentiometer position change / further) debouncing could be introduced here
    MIDI.sendControlChange(faderAssign[faderpage][a], fader[a], faderChout[faderpage][a]);
    faderValue[faderpage][a] = fader[a];
    }
}


void faderpage_select(){
  
    if (bitRead(eightbuttons,a) == HIGH && menu == 6 && digitalRead(functionbutton)==LOW) {
      
      faderpage = a;
         
      for (b=0; b <= 7; b++){faderAssignState[b] = 0;     d[b] = 0;}
      faderAssignState[a] = 1; //only for led output
      c = a;
    }
}


//receiving the eight buttons via shift register 74HC165 (per SPI)
void input_eightbuttons(){
  //digitalWrite(ss, 1); //74HC165: load into ShiftRegister
  eightbuttons = SPI.transfer(0x00);  //send byte (8bit) from 74HC165 (uses sck as clock and miso as input line)
  //digitalWrite(ss, 0);  //74HC165: inputs enabled
}


//sending ctrl changes (i.e. mute) via midi
void output_eightbuttons(){

  if (bitRead(eightbuttons,a) != bitRead(eightbuttonsLast,a) && t1-t2>5 && digitalRead(functionbutton)==LOW) { //button debouncing & only if function-pushbutton is not pressed
    eightbuttonsLast = eightbuttons; //saves last state of the eight buttons for button debouncing
    
    if (bitRead(eightbuttons,a) == HIGH && menu < 6) { //if current button of the eightbuttons = high
      buttonBool[menu][a] = !buttonBool[menu][a]; //alternates 'boolean type' buttonBool between 0 (i.e.=unmuted) to 1 (i.e.=muted)
      if (buttonBool[menu][a] == HIGH){
        MIDI.sendControlChange(buttonAssign[menu][a], buttonHigh[menu][a], buttonChout[menu][a]); // sends via midictrl cc#
      }
      if (buttonBool[menu][a] == LOW) {
        MIDI.sendControlChange(buttonAssign[menu][a], buttonLow[menu][a], buttonChout[menu][a]); // sends via midictrl cc#
      }
    }

    t2=millis(); //t2 for button debouncing
  }
}


//shows status on the 8 LEDs via output shiftregister 74HC595 (using SPI)
void output_ledrow(){
  //digitalWrite(ss, 1); //74HC595: load into Outputs
  
  if (digitalRead(functionbutton)==HIGH){ //if functionpushbutton is pressed, current mode is written into "ledrow"
    if(t1-t2>400){     //leds blinking
      switch(menu){
          case 0: ledrow = B00000001; break; //led #1 (LSB) is blinking when in mute mode 
          case 1: ledrow = B00000010; break;
          case 2: ledrow = B00000100; break;
          case 3: ledrow = B00001000; break;
          case 4: ledrow = B00010000; break;
          case 5: ledrow = B00100000; break;
          case 6: ledrow = B01000000; break;
      }
      t2=millis(); }
    if(t1-t2>200){ledrow = B00000000;} //for led blinking
  }
  
  if (digitalRead(functionbutton)==LOW) { //if function-puhbutton is not pressed, current channel status of current mode is written into "ledrow"
    if(menu < 6){
      if (buttonAssign[menu][a] == 49) {bitWrite(ledrow, a, !buttonBool[menu][a]);}
      else {bitWrite(ledrow, a, buttonBool[menu][a]);}
    }
    else {bitWrite(ledrow, a, faderAssignState[a]);}
  } 

  SPI.transfer(ledrow); //send byte (8bit) to the 8 LEDs
  //digitalWrite(ss, 0); //disable communication for 74HC165 (no change for 74HC595)
}

//receiving the eight buttons via input shift register 74HC165 (per SPI) and change button functionalities (i.e. mute, solo, recarm, ...)
void menu_eightbuttons(){
  
  if (bitRead(eightbuttons,a) != bitRead(eightbuttonsLast,a) && t1-t2>5 && bitRead(eightbuttons,a)==HIGH && digitalRead(functionbutton)==HIGH) { //button debouncing & only if function-pushbutton is pressed
    eightbuttonsLast = eightbuttons; //saves last state of the eight buttons for button debouncing

    if (a < 7){
      if(menu == a) { //undo all buttons on "one page" (a.k.a. "one menu")
        for (b=0; b <= 7; b++){
          buttonBool[menu][b] = 0;
          MIDI.sendControlChange(buttonAssign[menu][b], buttonLow[menu][b], buttonChout[menu][b]); }
      }
      if(menu != a){ //change button "page" (a.k.a. "menu")       
        menu = a;
      }

    }

    if (a == 7){
      MIDI.sendControlChange(61, 0, 1);
      
    }
    
    t2=millis(); //t2 for button debouncing
  }
}
    

void setup_menu_eightbuttons(){

  if (bitRead(eightbuttons,a) != bitRead(eightbuttonsLast,a) && t1-t2>5 && bitRead(eightbuttons,a)==HIGH && digitalRead(functionbutton)==HIGH) { //button debouncing & only if function-pushbutton is pressed
    eightbuttonsLast = eightbuttons; //saves last state of the eight buttons for button debouncing
    
    switch(a){
      case 0: menu = 0; break;
      case 1: menu = 1; break;
      case 2: menu = 2; break;
      case 3: menu = 3; break;
      case 4: menu = 4; break;
      case 5: menu = 5; break;
      case 6: menu = 6; break;
      case 7: menu = 7; break;
    }
    
    t2=millis(); //t2 for button debouncing
  }
}



//the arraypointer - used in several functions
void arraypointer(){
  a++; //increase pointer
  if(a > 7) { a = 0; } //reset pointer
}



//******************************************************************************************************************************
